/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


/*************************************************************************
 *
 * File : bt_node.c
 *
 * Modules:
 * Routines to mannipulate the B+ tree nodes
 *   BT_NewINode()
 *   BT_NewLNode()
 *   BT_NewONode()
 *   BT_NewLEntry()
 *   BT_InsertLNode()
 *   BT_InsertINode()   
 *   BT_InsertONode()   
 *   BT_SelectChild()
 *   BT_LNodeSplit()
 *   BT_INodeSplit()
 *   BT_InsertNewINode()
 *   BT_PrintLNode()
 *   BT_PrintINode()
 *   BT_VerifyLNode()
 *   BT_VerifyINode()
 *
 * Version 1.1
 * Coded by: Dai Haoyu		Modified by: Rui Zhang
 *************************************************************************/
#include "btree.h"
#include "bt_io.h"
#include "bt_node.h"

int max_top;

/***************************************************************************

* BT_NewLNode():
*   Routine to allocate memory a LNODE
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if memory is allocated successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_NewLNode(LNODE **node   /* pointer to the LNODE pointer */
		) {
  *node = (LNODE *)malloc(PAGE_SIZE);
  (**node).NumEntry = 0;
  (**node).Next = -1L;
  (**node).Overflow = -1L;   
  (**node).overflowNum = 0; 
  return (TRUE);
}

/***************************************************************************

* BT_NewONode():
*   Routine to allocate memory a ONODE
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if memory is allocated successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_NewONode(ONODE **node   /* pointer to the ONODE pointer */
		) {

  *node = (ONODE *)malloc(O_PAGE_SIZE);
  (**node).NumEntry = 0;
  (**node).Next = -1L;
  
  return (TRUE);
}

/***************************************************************************

* BT_NewINode():
*   Routine to allocate memory a INODE
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if memory is allocated successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_NewINode(INODE **node   /* pointer to the LNODE pointer */
		) {
  *node = (INODE *)malloc(PAGE_SIZE);
  (**node).NumKey = 0;

  return (TRUE);
}

/***************************************************************************

* BT_NewLEntry():
*   Routine to allocate memory space for a LENTRY
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if memory is allocated successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_NewLEntry(LENTRY **lentry
		 ) {
  *lentry = (LENTRY *)malloc(sizeof(LENTRY));
  /*  (**lentry).Overflow = -1L;*/

  return TRUE;
}

/***************************************************************************

* BT_InsertLNode():
*   Routine to add a new LENTRY to a LNODE
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if no split, FALSE if split occurs and the new LNODE resulting from
*   splitting will be passed through the last argument
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The last argument passes back the new node after splitting if any
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_InsertLNode(BTREE *btree,
                   LENTRY *entry,  /* the LENTRY to be added */
                   LNODE *lnode,   /* the LNODE to which to add the LENTRY */
                   LNODE **newLnode /* the new LNODE after splitting */
		   ) {
  int i, k;
  int found = FALSE;
  int position = 0 ;
  ONODE *onode = NULL;

  if ( lnode->NumEntry == L_FAN_OUT 
       && entry->Key == lnode->Entries[lnode->NumEntry-1].Key) {      
     
    BT_NewONode(&onode);
    if (lnode->Overflow >= 0) {
      BT_IOReadONode(btree, lnode->Overflow, onode);         
      if (BT_InsertONode(btree, entry, onode)){
	BT_IOWriteONode(btree, lnode->Overflow, onode);
      }
    } else {
      if (BT_InsertONode(btree, entry, onode)){
	lnode->Overflow = BT_IOWriteNewONode(btree, onode);
      }         
    }  
    lnode->overflowNum++;  

    free(onode);
    return (TRUE);      
  }   
  
  if ( lnode->NumEntry == L_FAN_OUT 
       && entry->Key < lnode->Entries[lnode->NumEntry-1].Key) {
     
    i = lnode->NumEntry-1;
    if (lnode->Entries[i].Key == lnode->Entries[i-1].Key) {
      BT_NewONode(&onode);
      if (lnode->Overflow >= 0) {
	BT_IOReadONode(btree, lnode->Overflow, onode);         
	if (BT_InsertONode(btree, &(lnode->Entries[i]), onode)){
	  BT_IOWriteONode(btree, lnode->Overflow, onode);
	}
      } else {         
	if (BT_InsertONode(btree, &(lnode->Entries[i]), onode)){
	  lnode->Overflow = BT_IOWriteNewONode(btree, onode);
	}
      }  
      lnode->overflowNum++;    

      free(onode);      
      lnode->NumEntry--;                    
    }
  } else {
  }
  
  if (lnode->NumEntry < L_FAN_OUT) { /* node is not full */
    i = 0;
    while (!found && i<lnode->NumEntry){
      if (entry->Key >= lnode->Entries[i].Key){
        i++;
      } else {
        found = TRUE;
      }   
    }     
    position = i;  
    
    for (i=lnode->NumEntry; i > position; i--){
      lnode->Entries[i].Key = lnode->Entries[i-1].Key;
      lnode->Entries[i].Offset = lnode->Entries[i-1].Offset;      
      for(k=0; k<D; k++){
	(lnode->Entries[i]).ver[k]=(lnode->Entries[i-1]).ver[k];
      }
      
    } 
    lnode->Entries[position].Key = entry->Key;
    lnode->Entries[position].Offset = entry->Offset;
    for(k=0; k<D; k++){       
      (lnode->Entries[position]).ver[k]=entry->ver[k];
    }    
    lnode->NumEntry += 1;
    return (TRUE);
  } else { /* node full, split needed */     
    BT_LNodeSplit(btree, entry, lnode, newLnode);
    return (FALSE);
  }
}

/***************************************************************************

* BT_InsertINode():
*   Routine to add a new entry to a INODE
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if no split, FALSE if split occurs and the new LNODE resulting from
*   splitting will be passed through the second last argument and the key 
*   value to be moved up is passed through the last argument
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The second last argument passes back the new node after splitting if any
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_InsertINode(PTR offset,  /* the offset of the key to be inserted */
                   KEYTYPE key, /* the key to be inserted */
                   PTR child,   /* the offset of the node which causes
                                 * this insertion */
                   INODE *node, /* the node to which to insert */
                   INODE **newNode, /* the new node if split occurs */
                   KEYTYPE *splitKey /* the key that is to be moved up */
		   ) {
  int position;
  int i;
  int found = FALSE;

  if (node->NumKey < I_FAN_OUT-1){  /* no split */
    for (i=0; i<=node->NumKey; i++){
      if ((node->Children[i] == child) && !found) {
	position = i;
	found = TRUE;
	break;
      }
    }
    for (i=node->NumKey; i>position; i--) {
      node->Keys[i] = node->Keys[i-1];
    }
    node->Keys[position] = key;
    for (i=node->NumKey+1; i>position+1; i--) {
      node->Children[i] = node->Children[i-1];
    }
    node->Children[position+1] = offset;
    node->NumKey += 1;
    
    return (TRUE);
  } else {  /* split needed */
    BT_INodeSplit(offset, key, child, node, newNode, splitKey);
    return (FALSE);
  }
}  

/***************************************************************************
 * BT_InsertONode():
 *   Routine to insert aan entry to a ONODE
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   TRUE if entry is inserted to the current node else FALSE
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   New overflow node may be allocated if the current node is full
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

int BT_InsertONode(BTREE *btree,
                   LENTRY *entry,
                   ONODE *node
		   ) {
  int k;
  ONODE *nextNode = NULL;

  if (node->NumEntry < O_FAN_OUT) {
    (node->Entries[node->NumEntry]).Key=entry->Key;
    for(k=0; k<D; k++){
      (node->Entries[node->NumEntry]).ver[k]=entry->ver[k];
    }    
    node->Entries[node->NumEntry].Offset = entry->Offset;
    (node->NumEntry)++;
    return (TRUE);
  } else { /* current nodde is full */
    if (node->Next >= 0L) { /* the next overflow page exists */
      BT_NewONode(&nextNode);
      BT_IOReadONode(btree, node->Next, nextNode);
      if (BT_InsertONode(btree, entry, nextNode)) {
	BT_IOWriteONode(btree, node->Next, nextNode);
      }
      free(nextNode);
      return (FALSE);
    } else { /* no next overflow page */
      BT_NewONode(&nextNode);
      if (BT_InsertONode(btree, entry, nextNode)){
	node->Next = BT_IOWriteNewONode(btree, nextNode);
      }
      free(nextNode);
      return (TRUE);
    }  
  }
}

/***************************************************************************

* BT_SelectChild():
*   Routine to select a child according to the key value s.t. the node 
*   pointed by the child would contain the key value
*   It will choose the first child it hits even there might be more 
*   qualifying children
*
* IMPORTS:
*   None.
*
* RETURNS:
*   The offset of the selected child node.
*   If the node is empty -1 will be returned
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
PTR BT_SelectChild(INODE *node, 
                   KEYTYPE key
		   ) {
  int i;

  if (node->NumKey == 0) {
    return (-1L);
  } else {
    if (key < node->Keys[0]) {
      return (node->Children[0]);
    }
    for (i=0; i<node->NumKey-1; i++){
      if ((key >= node->Keys[i] && key<node->Keys[i+1]) ||
          (key == node->Keys[i] && key<=node->Keys[i+1])){
         
        return (node->Children[i+1]);
      }
    }
    return (node->Children[node->NumKey]);
  }
}

/***************************************************************************

* BT_LNodeSplit():
*   ShortDescription
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE id splitting is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   Memory space will be allocated for the new node
*
* WARNINGS:
*   The first node should be full before insertion
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_LNodeSplit(BTREE *btree,
                  LENTRY *entry, /* the LENTRY to be inserted */
                  LNODE *node,   /* the full LNODE */
                  LNODE **newNode  /* the new node */
		  ) {
  int i = 0;
  int k,j;
  int position;
  PTR over;
  int found = FALSE;
  int total, sub_total;
  int splitP=-1;
  int overExist=0;
  ONODE *onode=NULL;
  ONODE *newOnode=NULL;
  LENTRY allEntry[L_FAN_OUT+1];
  KEYTYPE key;

  if (node->NumEntry != L_FAN_OUT) {
    printf("BT_LNodeSplit : node is not full \n");
    return (FALSE);
  }
  BT_NewLNode(newNode);
  
  while (!found && i<node->NumEntry){
    if (entry->Key >= node->Entries[i].Key){
      i++;
    } else {
      found = TRUE;
    } 
  }
  position = i; 
  
  if ( position == node->NumEntry) {
    key=(node->Entries[node->NumEntry-1]).Key;
    for (i=1; i<node->NumEntry; i++) {
      if ((node->Entries[node->NumEntry-1-i]).Key == key ) {           
      } else {           
	break;
      }
    }     
    position += node->overflowNum;
    if ((i+ node->overflowNum)>= L_FAN_OUT) {                         
      (**newNode).Entries[0].Key = entry->Key;
      (**newNode).Entries[0].Offset = entry->Offset;
      for(k=0; k<D; k++){
	((**newNode).Entries[0]).ver[k]=entry->ver[k];
      }   
      (**newNode).NumEntry = 1;         
      (**newNode).Overflow = -1;    
      (**newNode).overflowNum = 0;           
      return (TRUE);                   
    } else if ((i+ node->overflowNum)>= L_FAN_OUT/2) {
      splitP = node->NumEntry-i;      
      for (i=splitP; i<node->NumEntry; i++){
	(**newNode).Entries[i-splitP].Key = node->Entries[i].Key;
	(**newNode).Entries[i-splitP].Offset = node->Entries[i].Offset;
	for(k=0; k<D; k++){
	  ((**newNode).Entries[i-splitP]).ver[k] = node->Entries[i].ver[k];
	}          
      } 
 
      node->NumEntry = splitP;  
      i=i-splitP;
      if (node->overflowNum>0) {
	BT_NewONode(&onode);
	BT_IOReadONode(btree,node->Overflow,onode);
	for (j=0;j<onode->NumEntry;j++) {
	  (**newNode).Entries[i].Key = onode->Entries[j].Key;
	  (**newNode).Entries[i].Offset = onode->Entries[j].Offset;
	  for(k=0; k<D; k++){
	    ((**newNode).Entries[i]).ver[k] = onode->Entries[j].ver[k];
	  }                         
	  i++;
	}                                
	/*put the used overflow node address into trash*/
	(btree->top)++;            
	if ((btree->top) > max_top) {
	  max_top = btree->top;
	}
	if (max_top > MAX_STACK) {
	  printf("Stack size is too small !!\n");
	}
	btree->Stacks[btree->top]=node->Overflow;                        
	node->Overflow=-1;        
	node->overflowNum=0;
	free(onode);
      }
      (**newNode).NumEntry = i;                
      (**newNode).Entries[i].Key = entry->Key;
      (**newNode).Entries[i].Offset = entry->Offset;
      for(k=0; k<D; k++){
	(**newNode).Entries[i].ver[k]=entry->ver[k];
      }
      (**newNode).NumEntry++;  
      (**newNode).Overflow=-1;  
      (**newNode).overflowNum=0;  
      return (TRUE);                 
    } else {        
    }
  } else {
  } 
   
  total = node->NumEntry + 1 + node->overflowNum;   
  if (total > L_FAN_OUT+1) {
    sub_total = L_FAN_OUT+1;
  } else {
    sub_total=total;
  }
  splitP = total/2; 
  if (splitP >=  node->NumEntry) {
    splitP = node->NumEntry-1;
  }
  i=0;
  while ((node->Entries[splitP]).Key == (node->Entries[splitP-1]).Key) {
    i++;
    if (splitP+i < node->NumEntry 
	&& (node->Entries[splitP+i]).Key != (node->Entries[splitP+i-1]).Key) {
     
      splitP=splitP+i;
      break;
    } else if (splitP-i>0 
	       && (node->Entries[splitP-i]).Key != (node->Entries[splitP-i-1]).Key) {
     
      splitP=splitP-i;     
      break;
    }      
  }
  /*   printf("* %d^",splitP); */

  if (total > sub_total) { 
    BT_NewONode(&onode);
    BT_NewONode(&newOnode);
    overExist=1;
    over=node->Overflow;     
  }
     
  for (i=0; i<sub_total; i++){      
    if ( i<position && i<node->NumEntry){
      allEntry[i].Key = node->Entries[i].Key;
      allEntry[i].Offset = node->Entries[i].Offset;
      for(k=0; k<D; k++){       
	allEntry[i].ver[k]=(node->Entries[i]).ver[k];
      }      
  
    } else if (i==position){
      allEntry[i].Key = entry->Key;
      allEntry[i].Offset = entry->Offset;      
      for(k=0; k<D; k++){
	allEntry[i].ver[k]=entry->ver[k];
      }
       
    } else  {      
      allEntry[i].Key = node->Entries[i-1].Key;
      allEntry[i].Offset = node->Entries[i-1].Offset;
      for(k=0; k<D; k++){
	allEntry[i].ver[k]=(node->Entries[i-1]).ver[k];
      } 
    } 
  }  
     
  (**newNode).NumEntry=0;
  (**newNode).overflowNum=0;
  for (i=0; i<sub_total; i++){    
    if (i<splitP){
      node->Entries[i].Key = allEntry[i].Key;
      node->Entries[i].Offset = allEntry[i].Offset;
      for(k=0; k<D; k++){
	(node->Entries[i]).ver[k]=allEntry[i].ver[k];
      }      
    } else {           
      (**newNode).Entries[i-splitP].Key = allEntry[i].Key;
      (**newNode).Entries[i-splitP].Offset = allEntry[i].Offset;
      for(k=0; k<D; k++){
	((**newNode).Entries[i-splitP]).ver[k]=allEntry[i].ver[k];
      }     
      (**newNode).NumEntry++; 
    }
  }   
  i=i-splitP;
  if (overExist==1){  
    overExist=0;    
    while (over>=0) { 
      BT_IOReadONode(btree,over,onode);
      for (j=0;j<onode->NumEntry;j++) {
	if ((**newNode).NumEntry < L_FAN_OUT) {
	  (**newNode).Entries[i].Key = onode->Entries[j].Key;
	  (**newNode).Entries[i].Offset = onode->Entries[j].Offset;
	  for (k=0; k<D; k++){
	    (**newNode).Entries[i].ver[k] = onode->Entries[j].ver[k];
	  } 
	  i++; 
	  (**newNode).NumEntry++; 
	} else {
	  overExist=1;
	  BT_InsertONode(btree,&(onode->Entries[j]),newOnode);        
	  (**newNode).overflowNum++;                     
	}  
      }               
      /*put the used overflow node address into trash*/     
      (btree->top)++;
      if ((btree->top) > max_top) {
	max_top = btree->top;
      }          
      if (max_top > MAX_STACK) {
	printf("Stack size is too small !!\n");               
      }             
      btree->Stacks[btree->top]=over;                   
        
      over=onode->Next;  
    }               
    node->Overflow=-1;           
    node->overflowNum=0;                                
    node->NumEntry = splitP;
    if (overExist==1) {
      (**newNode).Overflow=BT_IOWriteNewONode(btree,newOnode);
    }   
    free(onode);     
    free(newOnode);     
    return (TRUE);     
  } else { 
    node->NumEntry = splitP;     
    return (TRUE);         
  } 
}

/***************************************************************************
 * BT_INodeSplit():
 *   Routine to perform splitting of INODE
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   Return TRUE if splitting is successful else FALSE
 *   It also passes back the key which is to be moved up 
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   Memory space will be allocated for the new node
 *
 * WARNINGS:
 *   The first node should be full before insertion
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_INodeSplit(PTR offset,
                  KEYTYPE key,
                  PTR child,
                  INODE *node,
                  INODE **newNode,
                  KEYTYPE *splitKey
		  ) {
  KEYTYPE allKey[I_FAN_OUT];
  PTR allChildren[I_FAN_OUT+1];
  int i;
  int totalKey;
  int totalChildren;
  int position;
  int splitP;
  int found = FALSE;

  if (node->NumKey != (I_FAN_OUT-1)){
    printf("BT_INodeSplit : node is not full \n");
  }
  BT_NewINode(newNode);

  for (i=0; i<=node->NumKey; i++){
    if ((node->Children[i] == child) && !found) {
      position = i;
      found = TRUE;
    }
  }
  totalKey = I_FAN_OUT;
  totalChildren = totalKey + 1;
  splitP = totalKey/2;

  for (i=0; i<totalKey; i++){
    if (i<position){
      allKey[i] = node->Keys[i];
    } else 
      if (i == position) {
	allKey[i] = key;
      } else {
	allKey[i] = node->Keys[i-1];
      }
  }

  for (i=0; i<totalChildren; i++){
    if (i<position+1){
      allChildren[i] = node->Children[i];
    } else 
      if (i == position+1) {
	allChildren[i] = offset;
      } else {
	allChildren[i] = node->Children[i-1];
      }
  }

  for (i=0; i<totalKey; i++) {
    if (i < splitP) {
      node->Keys[i] = allKey[i];
    } else if (i == splitP) {
    } else {
      (**newNode).Keys[i-splitP-1] = allKey[i];
    }
  }

  for (i=0; i<totalChildren; i++) {
    if (i <= splitP) {
      node->Children[i] = allChildren[i];
    } else {
      (**newNode).Children[i-splitP-1] = allChildren[i];
    }
  }
  node->NumKey = splitP;
  (**newNode).NumKey = totalKey - splitP - 1;
  (**newNode).Level = node->Level;
  *splitKey = allKey[splitP];

  return (TRUE);
}

/***************************************************************************
 * BT_InsertNewINode():
 *   Routine to insert entries to an empty INODE
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   TRUE if insertion is successful else FALSE.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   The empty node will have two children and one key
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_InsertNewINode(INODE *node,   /*node to which insertion to be made */
                      PTR child1,    /*the offset of the first child */
                      KEYTYPE key,   /*the value of the first key */
                      PTR child2     /*the offset of the second child */
		      ) {
  node->Keys[0] = key;
  node->Children[0] = child1;
  node->Children[1] = child2;
  node->NumKey = 1;

  return (TRUE);
}

/***************************************************************************
 * BT_PrintLNode():
 *   Routine to print the content of a LNODE
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
void BT_PrintLNode(LNODE *node
		   ) {
  int i;
 
  printf("LNODE$$NumEntry : %d Next %d\n", node->NumEntry, node->Next);

  for (i=0; i<node->NumEntry; i++) {
    printf("Entry %d : ", i);
    printf("KEY: %.18f, RECID: %d \n  ", node->Entries[i].Key,
           node->Entries[i].Offset);    
  }
  printf("OVFL: %d \n", node->Overflow);
}

/***************************************************************************
 * BT_PrintINode():
 *   Routine to print the content of a INODE
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
void BT_PrintINode(INODE *node
		   ) {
  int i;
 
  printf("INODE##NumKey : %d Level %d\n", node->NumKey, node->Level);

  for (i=0; i<node->NumKey; i++) {
    printf("  Child %d : %d\n", i, node->Children[i]);
    printf("Key %d : %.18f\n", i, node->Keys[i]);
  }
  printf("        Child %d : %d\n", i, node->Children[i]);
  
}

/***************************************************************************
 * BT_VerifyLNode():
 *   Routine to verify the constraints of B+ tree for a leaf node
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   TRUE if the node is in correct state else FALSE
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_VerifyLNode(LNODE *lnode
		   ) {
  int i;

  for (i=0; i<lnode->NumEntry-1; i++) {
    if (lnode->Entries[i].Key <= lnode->Entries[i+1].Key) {
    } else {
      printf("BT_VerifyLNode : entries not in correct order\n");
      return (FALSE);
    }
  }
  
  return (TRUE);
}

/***************************************************************************
 * BT_VerifyINode():
 *   Routine to verify the constraints of B+ tree for a internal node
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   TRUE if the node is in correct state else FALSE
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_VerifyINode(INODE *inode
		   ) {
  int i;

  for (i=0; i<inode->NumKey-1; i++) {
    if (inode->Keys[i] <= inode->Keys[i+1]) {
    } else {
      printf("BT_VerifyINode : entries not in correct order\n");
      return (FALSE);
    }
  }
  
  return (TRUE);
}
