/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

#ifndef __BTREE_H
#define __BTREE_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <malloc.h>
#include <float.h>

#define D 32

#define compactrate 0.9f //should be set larger than 0.75 or less than 0; if compactrate<0, no compacting.
#define dr 0.005f //delta r in kNN search
#define r0 0.05f //initial search radius in kNN search

typedef float DATA_TYPE;

//*
typedef float KEYTYPE;
#define KEY_MAX FLT_MAX
#define KEY_MIN FLT_MIN
//*/

/*
  typedef unsigned long KEYTYPE;
  #define KEY_MAX ULONG_MAX
  #define KEY_MIN 0
//*/

//typedef unsigned long long KEYTYPE;



#ifndef PAGE_SIZE
#define PAGE_SIZE 4096
#endif

#define O_PAGE_SIZE 4096

#ifndef MAX_STACK
#define MAX_STACK 300
#endif

#define BUF_SIZE 1

#define TRUE 1
#define FALSE 0

typedef long PTR;

typedef struct {
  DATA_TYPE fields[D];
} RECORD_TYPE;

typedef struct {
  DATA_TYPE lower[D];
  DATA_TYPE higher[D];
} RANGE_TYPE;

typedef struct {
  float start[D];   
  float end[D];   
} FLTRANGE; //this may be used to specify range queries; some old functions use this type.

#define I_FAN_OUT ((int)(PAGE_SIZE-(2*(sizeof(int)))+(sizeof(KEYTYPE)))/((sizeof(KEYTYPE))+(sizeof(PTR))))

typedef struct{
  int Level;
  int NumKey;
  KEYTYPE Keys[I_FAN_OUT-1];
  PTR     Children[I_FAN_OUT];
} INODE;

typedef struct _IMINODE{
  int Level;
  int NumKey;
  KEYTYPE Keys[I_FAN_OUT-1];
  struct _IMINODE *Children[I_FAN_OUT];
} IMINODE;

typedef struct{
  KEYTYPE Key;
  DATA_TYPE ver[D];
  long Offset;
} LENTRY;

#define L_FAN_OUT (((int)(PAGE_SIZE-(sizeof(int))*2-(sizeof(long))*3)/(sizeof(LENTRY))))

typedef struct{
  int NumEntry;
  PTR Next;  
  PTR Prev;
  LENTRY Entries[L_FAN_OUT];
  PTR Overflow;
  int overflowNum;
} LNODE;

#define O_FAN_OUT (((int)(O_PAGE_SIZE-(sizeof(int))-(sizeof(long)))/(sizeof(LENTRY))))

typedef struct{
  int NumEntry;
  PTR Next;
  LENTRY Entries[O_FAN_OUT];
} ONODE;

typedef struct{
  FILE* BTreeFile;  /* the file where the b+ tree is stored */
  FILE* SrhResult;  /* the file to store searching result form the btree */
  PTR   Root;       /* the offset of root node in the file  */
  PTR   Leaf_Head;  /* the offset of the first leaf node, from which 
                     * a sequencial scan can be started */
  PTR   Stacks[MAX_STACK];
  int   top;
  IMINODE *imroot;
} BTREE;

typedef struct{
  FILE* Fp;
  long  Offset;
}BUFFER_ENTRY;

typedef struct _OFFSETLIST{
  long Offset;
  struct _OFFSETLIST *next;
}OFFSETLIST;

typedef struct _RecordListNode{
  RECORD_TYPE rcd;
  long rid;
  struct _RecordListNode *next;
} RecordListNode;

typedef struct _IntervalListNode{
  KEYTYPE lower;
  KEYTYPE higher;
  struct _IntervalListNode *next;
} IntervalListNode;

typedef struct _Interval{
  KEYTYPE lower;
  KEYTYPE higher;
} Interval;

typedef struct _KnnInfo{
  int k;
  int numanswer;
  int farthestnum;
  DATA_TYPE farthest;
  DATA_TYPE *distance;
  DATA_TYPE r;
}KnnInfo;

int BT_Open(BTREE *, FILE *, PTR, PTR);
int BT_Close(BTREE *);
int BT_Insert(BTREE *, const long, const KEYTYPE, RECORD_TYPE *);
int BT_Search(BTREE *, PTR, RECORD_TYPE *, const KEYTYPE);
PTR BT_Locate(BTREE *, PTR, const KEYTYPE);
int BT_SearchRange(BTREE *, PTR, const KEYTYPE, const KEYTYPE, FLTRANGE *);
void refine(BTREE *);
void compact(BTREE *btree);
int BuildINodes(PTR ptr, BTREE *btree, KEYTYPE *keymin, KEYTYPE *keymax);
IMINODE* LoadINodes(PTR ptr, BTREE *btree);
PTR IMBT_Locate(IMINODE *, KEYTYPE);
IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key);
int BT_RangeSearch_SortedIntervals(BTREE *btree, RecordListNode *cdd, Interval *intvlarray, int ni);
int BT_RangeSearch_kNN(BTREE *btree, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo, IntervalListNode *intvlarray);
void SearchRight(PTR address, float curkeyhigh, BTREE *btree, int curpar, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo);
void SearchLeft(PTR address, float curkeylow, BTREE *btree, int curpar, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo);
void ProcessPoint(LENTRY le, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo);
int BT_DeleteKey(BTREE *btree, const KEYTYPE key);
int BT_DeleteRecord(BTREE *btree, RECORD_TYPE *rd, const KEYTYPE key);
#endif
