/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */



/*************************************************************************
 *
 * File : customize.c
 *
 * Modules:
 * Functions to customize the GiMP
 *	Reference()
 *	Distance()
 *	Base()
 *	MapRange()
 *	MapAnnulus()
 *	void Initialize()
 *	ActualDistance()
 *
 * Version 1.1
 * Coded by: Rui Zhang
 *************************************************************************/

#include "btree.h"
#include "support.h"
#include "customize.h"

//Definition of the functions

RECORD_TYPE* Reference (RECORD_TYPE *pRcd, int *refnum)
{
  RECORD_TYPE *Pr;
  int i, j;
  Pr = (RECORD_TYPE *) malloc (sizeof (RECORD_TYPE));

#ifdef BTREE_
  for (j = 0; j < D; j++)
    Pr->fields[j] = 0;
  *refnum = 0;
#endif

#ifdef IMINMAX_
  for (j = 0; j < D; j++)
    Pr->fields[j] = 0;
  *refnum = 0;
#endif

#ifdef UBTREE_
  for (j = 0; j < D; j++)
    Pr->fields[j] = 0;
  *refnum = 0;
#endif

#ifdef PYRAMID_
  for (j = 0; j < D; j++)
    Pr->fields[j] = 0.5;
  *refnum = (int) pv ( (DATA_TYPE *)pRcd);
#endif

#ifdef IDISTANCE_
  register DATA_TYPE temp;
  DATA_TYPE dist, mindist = KEY_MAX;
  int mindistnum = 0;

  for (i = 0; i < NumRef; i++) 
    {
      //calculate the distance to the ith ref point
      dist = 0;
      for (j = 0; j < D; j++)
	{
	  temp = pRcd->fields[j] - RefPoint[i][j];
	  dist += temp * temp;
	}

      //record it if better
      if (dist < mindist)
	{
	  mindist = dist;
	  mindistnum = i;
	}
    }

  for (j = 0; j < D; j++)
    Pr->fields[j] = RefPoint[mindistnum][j];
  *refnum = mindistnum;
#endif

  return Pr;
}

KEYTYPE Distance (RECORD_TYPE* pRcd1, RECORD_TYPE* pRcd2)
{

#ifdef BTREE_
  return pRcd1->fields[0] - pRcd2->fields[0];
#endif

#ifdef IMINMAX_
  int i, j;
  i = Dmax (pRcd1);
  j = Dmin (pRcd1);
  if (pRcd1->fields[j] < 1 - pRcd1->fields[i])
    return pRcd1->fields[j] - pRcd2->fields[j];
  else
    return pRcd1->fields[i] - pRcd2->fields[i];
#endif

#ifdef UBTREE_
  return ZV (pRcd1) - ZV (pRcd2); //when data are float point and normalized to [0,1]
#endif

#ifdef PYRAMID_
  int i;
  i = (int) pv ( (DATA_TYPE *)pRcd1);
  i = i % D;
  return fabs (pRcd1->fields[i] - pRcd2->fields[i]);
#endif

#ifdef IDISTANCE_
  int i;
  register DATA_TYPE temp;
  DATA_TYPE dist = 0;
  for (i = 0; i < D; i++) 
    {
      temp = pRcd1->fields[i] - pRcd2->fields[i];
      dist += temp * temp;
    }
  return sqrt (dist);
#endif
}

KEYTYPE Base (RECORD_TYPE *pRcd)
{
#ifdef BTREE_
  return 0;
#endif

#ifdef IMINMAX_
  int i, j;
  i = Dmax (pRcd);
  j = Dmin (pRcd);
  if (pRcd->fields[j] < 1 - pRcd->fields[i])
    return (KEYTYPE) j;
  else
    return (KEYTYPE) i;
#endif

#ifdef UBTREE_
  return 0;
#endif

#ifdef PYRAMID_
  return (int) pv ( (DATA_TYPE *)pRcd); //obtain pyramid number
#endif

#ifdef IDISTANCE_
  int refnum;
  Reference (pRcd, &refnum);
  return (KEYTYPE) refnum * (KEYTYPE) CST;
#endif
}

//map a range query to a set of one-dimensional intervals
//refer to the paper for more details
IntervalListNode* MapRange (RANGE_TYPE* pRg)
{
#ifdef BTREE_
  IntervalListNode *pILN;
  pILN = (IntervalListNode *) malloc (sizeof (IntervalListNode));
  pILN->lower = pRg->lower[0];
  pILN->higher = pRg->higher[0];
  pILN->next = NULL;
  return pILN;
#endif

#ifdef IMINMAX_
  IntervalListNode *pHead, *pILN;
  pILN = (IntervalListNode *) malloc (sizeof (IntervalListNode));
  pHead = pILN;

  int k, i1, j1, i2, j2;
  DATA_TYPE low, high;

  i1 = Dmax ( (RECORD_TYPE*) (pRg->lower));
  j1 = Dmin ( (RECORD_TYPE*) (pRg->lower));
  i2 = Dmax ( (RECORD_TYPE*) (pRg->higher));
  j2 = Dmin ( (RECORD_TYPE*) (pRg->higher));

  for (k = 0; k < D; k++)
    {
      if (pRg->lower[j1] >= 1 - pRg->lower[i1])
	{
	  low = k + pRg->lower[i1];
	  high = k + pRg->higher[k];
	}
      else if (pRg->higher[j2] < 1 - pRg->higher[i2])
	{
	  low = k + pRg->lower[k];
	  high = k + pRg->higher[j2];
	}
      else
	{
	  low = k + pRg->lower[k];
	  high = k + pRg->higher[k];
	}
      pILN->lower = low;
      pILN->higher = high;
      if (k < D-1)
	{
	  pILN->next = (IntervalListNode *) malloc (sizeof (IntervalListNode));
	  pILN = pILN->next;
	}
      else
	pILN->next = NULL;
    }
  return pHead;
	
#endif

#ifdef UBTREE_
  IntervalListNode *pHead, *pILN;
  int i;
  DATA_TYPE low, high;
  pHead = NULL;

  Z_TYPE start, end, cur;
  cur = start = ZV ( (RECORD_TYPE *) pRg->lower);
  end = ZV ( (RECORD_TYPE *) pRg->higher);

  while (cur <= end)
    {
      pILN = pHead;
      pHead = (IntervalListNode *) malloc (sizeof (IntervalListNode));
      pHead->next = pILN;
      cur = pHead->lower = getNextZvalue(cur, start, end);
      cur = pHead->higher = getNextZvalueExit(cur, start, end);
      if (cur == maxz)
	break;
      else
	cur++;
    }
  return pHead;
#endif

#ifdef PYRAMID_
  IntervalListNode *pHead, *pILN;
  int i;
  DATA_TYPE low, high;
  pHead=NULL;

  for (i = 0; i < 2 * D; i++)
    {
      if (intersect(pRg, i, &low, &high))
	{
	  pILN = pHead;
	  pHead = (IntervalListNode *) malloc (sizeof (IntervalListNode));
	  pHead->next = pILN;
	  pHead->lower = i+low;
	  pHead->higher = i+high;
	}
    }
  return pHead;
#endif
}


KEYTYPE prevkeylow[NumRef], prevkeyhigh[NumRef];
extern int newqueryflag;
extern DATA_TYPE min_r[NumRef], max_r[NumRef];

//map an annulus-shaped region to a set of one-dimensional intervals
//refer to the paper for more details
IntervalListNode* MapAnnulus (RECORD_TYPE *q, DATA_TYPE rmin, DATA_TYPE rmax, KnnInfo *knninfo)
{
  int i, j;
  IntervalListNode *pILN;
  KEYTYPE curkeylow, curkeyhigh;

#ifdef BTREE_
  pILN = (IntervalListNode *) malloc (sizeof (IntervalListNode) * 2 * NumRef);

  curkeylow = q->fields[0] - rmax;
  curkeyhigh = q->fields[0] + rmax;

  if (prevkeylow[0] < 0) //this partition has not been searched before
    {
      prevkeylow[0] = pILN[0].lower = curkeylow;
      pILN[0].higher = curkeyhigh;
      prevkeyhigh[0] = pILN[1].higher = curkeyhigh;
      pILN[1].lower = curkeyhigh;
    }
  else
    {
      if (prevkeylow[0] > min_r[0])
	{
	  pILN[0].higher = prevkeylow[0];
	  prevkeylow[0] = pILN[0].lower = curkeylow;
	}

      if (prevkeyhigh[0] < max_r[0])
	{
	  pILN[1].lower = prevkeyhigh[0];
	  prevkeyhigh[0] = pILN[1].higher = curkeyhigh;
	}
    }
#endif

#ifdef PYRAMID_
  RANGE_TYPE range;
  DATA_TYPE low, high;

  pILN = (IntervalListNode *) malloc (sizeof (IntervalListNode) * 2 * NumRef);

  for (i = 0; i < D; i++)
    {
      range.lower[i] = q->fields[i] - rmax;
      if (range.lower[i] < 0)
	range.lower[i] = 0;
      range.higher[i] = q->fields[i] + rmax;
      if (range.higher[i] > 1)
	range.higher[i] = 1;
    }

  for (i = 0; i < 2 * D; i++)
    {
      if (intersect (&range, i, &low, &high))
	{
	  curkeylow = i + low;
	  curkeyhigh = i + high;

	  if (prevkeylow[i] < 0) //this partition has not been searched before
	    {
	      prevkeylow[i] = pILN[i * 2].lower = curkeylow;
	      pILN[i * 2].higher = curkeyhigh;
	      prevkeyhigh[i] = pILN[i * 2 + 1].higher = curkeyhigh;
	      pILN[i * 2 + 1].lower = curkeyhigh;
	    }

	  else
	    {
	      if(prevkeylow[i] > min_r[i])
		{
		  pILN[i * 2].higher = prevkeylow[i];
		  prevkeylow[i] = pILN[i * 2].lower = curkeylow;
		}
	      else
		{
		  pILN[i * 2].higher = prevkeylow[i];
		  pILN[i * 2].lower = min_r[i];
		}

	      if(prevkeyhigh[i] < max_r[i])
		{
		  pILN[i * 2 + 1].lower = prevkeyhigh[i];
		  prevkeyhigh[i] = pILN[i * 2 + 1].higher = curkeyhigh;
		}
	      else
		{
		  pILN[i * 2 + 1].lower = prevkeyhigh[i];
		  pILN[i * 2 + 1].higher = max_r[i];
		}
	    }
	}
      else pILN[i * 2].lower = -1;
    }
#endif

#ifdef IDISTANCE_
  static DATA_TYPE qrdist[NumRef]; //distance between query and reference points

  pILN = (IntervalListNode *) malloc (sizeof (IntervalListNode) * 2 * NumRef);

  for (i = 0; i < NumRef; i++)
    {
      if (newqueryflag)
	qrdist[i] = ActualDistance (q, (RECORD_TYPE *) RefPoint[i]) + i * CST;

      if (qrdist[i] - i * CST <= knninfo->r + max_r[i])
	{
	  curkeylow = qrdist[i] - knninfo->r;
	  if (curkeylow < min_r[i])
	    curkeylow = min_r[i];

	  curkeyhigh = qrdist[i] + knninfo->r;
	  if (curkeyhigh > max_r[i])
	    curkeyhigh = max_r[i];

	  if (prevkeylow[i] < 0) //this partition has not been searched before
	    {
	      prevkeylow[i] = pILN[i * 2].lower = curkeylow;
	      pILN[i * 2].higher = curkeyhigh;
	      prevkeyhigh[i] = pILN[i * 2 + 1].higher = curkeyhigh;
	      pILN[i * 2 + 1].lower = curkeyhigh;
	    }
	  else
	    {
	      if (prevkeylow[i] > min_r[i])
		{
		  pILN[i * 2].higher = prevkeylow[i];
		  prevkeylow[i] = pILN[i * 2].lower = curkeylow;
		}
	      else
		{
		  pILN[i * 2].higher = prevkeylow[i];
		  pILN[i * 2].lower = min_r[i];
		}

	      if (prevkeyhigh[i] < max_r[i])
		{
		  pILN[i * 2 + 1].lower = prevkeyhigh[i];
		  prevkeyhigh[i] = pILN[i * 2 + 1].higher = curkeyhigh;
		}
	      else
		{
		  pILN[i * 2 + 1].lower = prevkeyhigh[i];
		  pILN[i * 2 + 1].higher = max_r[i];
		}
	    }
	}
      else pILN[i * 2].lower = -1;
    }

  for (i = 0; i < 2 * NumRef - 1; i++)
    pILN[i].next = &pILN[i + 1];
  pILN[2 * NumRef - 1].next = NULL;
#endif

  return pILN;
}

void Initialize ()
{
  int i, j;

#ifdef UBTREE_
  Z_Initialize ();
#endif

#ifdef IDISTANCE_
  if ( (fp_reference = fopen ("reference", "r")) == NULL)
    {
      printf ("Can not open file reference.\n");
      exit(0);
    }

  for (i = 0; i < NumRef; i++) 
    {
      for (j = 0; j < D; j++)
	fscanf (fp_reference, "%f", &(RefPoint[i][j]));
    }

  fclose (fp_reference);
#endif
}

DATA_TYPE ActualDistance (RECORD_TYPE *p1,RECORD_TYPE *p2)
{
  int j;
  register DATA_TYPE temp;
  DATA_TYPE dist = 0;
  for (j = 0; j < D; j++)
    {
      temp = p1->fields[j] - p2->fields[j];
      dist += temp * temp;
    }
  return (DATA_TYPE) sqrt ( (double)dist);
}
