/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */



/*************************************************************************
 *
 * File : gimp.c
 * Modules:
 * The main function
 *	main()
 *
 * Version 1.1
 * Coded by: Rui Zhang
 *************************************************************************/

#include "btree.h"
#include "nulllist.h"
#include "customize.h"
#include "operations.h"
#include "support.h"

main (int argc, char ** argv)
{

  extern long PageAccess;
  extern int QueueSize, max_top, newqueryflag;
  extern DATA_TYPE min_r[NumRef], max_r[NumRef];
  extern KEYTYPE keylow[NumRef], keyhigh[NumRef], prevkeylow[NumRef], prevkeyhigh[NumRef];
  extern PTR pagelow[NumRef], pagehigh[NumRef], donelow[NumRef+1], donehigh[NumRef+1];

  int i, curquery;
  int numquery = 10; //number of queries to process
  int k = 10; //k in kNN
  long totalpage;

  RANGE_TYPE rangequery;
  DATA_TYPE orirg[D][2], ar;
  RECORD_TYPE knnquery, rcd;
	
  KnnInfo knninfo;

  RecordListNode *pRcd; //Note: pRcd is the header of the answer list.
  //It points to the first answer point. Itself doesnot contain any answer.
	
  INODE *iNode;
  LNODE *lNode;
  ONODE *oNode;

  BTREE tree;
	
  FILE *fp_data;
  FILE *fp_index;
  FILE *fp_query;
  FILE *fp_knnquery;
  FILE *fp_btreeinfo;	
  FILE *fp_result;

  char DataFileName[50] = {"data"};
  char IndexFileName[50] = {"index"};
  char QueryFileName[50] = {"rangequery"};
  char kNNQueryFileName[50] = {"knnquery"};

  strcpy (DataFileName, "data"); //the data file

  strcpy (QueryFileName, "rangequery"); //range query file

  strcpy (kNNQueryFileName, "knnquery"); //knn query file

  QueueSize = 0;
  max_top = -1;
  Initialize ();
	
  if (argc == 2 && strcmp (argv[1], "b") == 0) //build index
    {
      for (i = 0; i < NumRef; i++)
	{
	  min_r[i] = FLT_MAX;
	  max_r[i] = 0.0;
	}

      //build index, function defined in operations.c
      BuildIndex(&tree, DataFileName, fp_data, IndexFileName, fp_index, fp_btreeinfo);
    }

  else if (argc == 2 && (strcmp (argv[1], "r") == 0 || strcmp (argv[1], "n") == 0)) //range query or knn query
    {
      if ( (fp_knnquery = fopen (kNNQueryFileName, "r")) == NULL) //open KNNQueryFile
	{
	  printf ("Can not open file %s\n", kNNQueryFileName);
	  exit (0);
	}

      if ( (fp_query = fopen (QueryFileName, "r")) == NULL) //open rangeQueryFile
	{
	  printf ("Can not open file %s\n", QueryFileName);
	  exit (0);
	}

      if ( (fp_index = fopen (IndexFileName, "r+")) == NULL) //open indexFile
	{
	  printf ("Can not open file %s\n", IndexFileName);
	  exit (0);
	}

      if ( (fp_btreeinfo = fopen ("btreeinfo", "r")) == NULL) //open btreeInfo
	{
	  printf ("Can not open file btreeinfo\n");
	  exit (0);
	}

      if ( (fp_result = fopen("result", "w+")) == NULL) //open resultFile to write
	{
	  printf ("Can not open file result\n");
	  exit (0);
	}

      fscanf (fp_btreeinfo, "%ld", &(tree.Root));
      fscanf (fp_btreeinfo, "%ld", &(tree.Leaf_Head));

      //open a B+ tree, function defined in btree.c
      BT_Open (&tree, fp_index, tree.Root, tree.Leaf_Head);

      curquery = 0;
      totalpage = 0;

      //load internal node, defined in btree.c
      tree.imroot = LoadINodes (tree.Root, &tree);

      if (strcmp (argv[1], "r") == 0) //range query
	{
	  while (fread (orirg, sizeof (RANGE_TYPE), 1, fp_query) && curquery < numquery)
	    {
	      PageAccess = 0;

	      //load range query
	      for (i = 0; i < D; i++)
		{
		  rangequery.lower[i] = orirg[i][0];
		  rangequery.higher[i] = orirg[i][1];
		}
	      fprintf (fp_result, "query %d\n", curquery);
	      for (i = 0; i < D; i++)
		fprintf (fp_result, "%f %f,", (float) orirg[i][0], (float) orirg[i][1]);
	      fprintf (fp_result, "\n");

	      //process range query, function defined in operations.c
	      pRcd = RangeSearch (&rangequery, &tree);

	      //output answers, function defined in operations.c
	      PrintAnswers (pRcd, fp_result);
	      //fprintf (fp_result,"Page access: %ld\n",PageAccess);
	      fprintf (fp_result, "\n");
	      totalpage += PageAccess;
	      curquery++;
	    }
	  fprintf (fp_result, "Average page access: %f\n", (float) totalpage / (float) curquery);
	  printf("Range search done. Result is in file \"result\".\n");
	}

      else if (strcmp (argv[1], "n") == 0) //KNN query
	{
	  for (i = 0; i < NumRef; i++) 
	    {
	      fscanf (fp_btreeinfo, "%f ", &min_r[i]);
	      fscanf (fp_btreeinfo, "%f\n", &max_r[i]);   
	    }

	  ar = 0;

	  while (fread (&knnquery, sizeof(RECORD_TYPE), 1, fp_knnquery) && curquery < numquery)
	    {
	      //initializations for each query
	      PageAccess = 0;
	      for (i = 0; i < NumRef; i++)
		prevkeylow[i] = prevkeyhigh[i] = keylow[i] = keyhigh[i] = pagelow[i] = pagehigh[i] = donelow[i] = donehigh[i] = -1;
	      donelow[NumRef] = donehigh[NumRef] = -1;
	      newqueryflag = 1;
	      //End initializations for each query

	      fprintf (fp_result, "query %d\n", curquery);
	      for (i = 0; i < D; i++)
		fprintf (fp_result, "%f ", (float) knnquery.fields[i]);
	      fprintf (fp_result, "\n");

	      //construct knn query
	      knninfo.distance = (DATA_TYPE *) malloc (sizeof (DATA_TYPE) * k);
	      knninfo.k = k;
	      knninfo.numanswer = knninfo.farthestnum = knninfo.farthest = 0;
	      //construction finished

	      //process KNN query, function defined in operations.c
	      pRcd = kNNSearch (&knnquery, &tree, &knninfo);

	      //output answers, function defined in operations.c
	      PrintAnswers (pRcd, fp_result);
	      //fprintf (fp_result,"Page access: %ld\n",PageAccess);
	      //fprintf (fp_result,"\n");

	      ar += sqrt(knninfo.farthest);
	      totalpage += PageAccess;
	      curquery++;
	    }
	  fprintf (fp_result, "Average page access: %f\n", (float) totalpage / (float) curquery);
	  fprintf (fp_result, "Average KNN distance: %f\n", ar / (float) curquery);
	  printf("KNN search done. Result is in file \"result\".\n");
	}

      //close all files
      fclose (fp_knnquery);
      fclose (fp_query);
      fclose (fp_btreeinfo);
      fclose (fp_index);
      fclose (fp_result);
    }

  else printf ("please enter a choice:\n b (build index)\n r (range query)\n n (knn query)\n");
}
