/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


 
/*************************************************************************
 *
 * File : io.c
 *
 * Modules:
 * Routines to perform basic IO functions
 *	IO_PageAccess()
 *	IO_DisplayPageAccess()
 *	IO_DisplayPageAccess_v2()
 *	IO_ResetPageAccess()
 *	IO_ResetPageAccess_v2()
 *	IOFileExist()
 *	IOOpen()
 *	IOClose()
 *
 * Version 1.1
 * Coded by: Dai Haoyu
 *************************************************************************/
#include "btree.h"
#include "io.h"

/***************************************************************************

* IO_PageAccess():
*   Routine to record a page access issued by the program. If the page
*   is already in the buffer, PageAccess is not increased. Else the page will
*   be brought into the buffer and it may cause the least recently accessed
*   page to be flushed if there is no free space in the memory.
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void  IO_PageAccess(FILE *fp,
                    long offset
		    ) {
  int i, j, tQS;
  int found = FALSE;

  if (offset < 0L)
    return;

  for (i=0; i<QueueSize; i++) {
    if (Buffer[i].Fp == fp && Buffer[i].Offset == offset) {
      found = TRUE;
      break;
    }
  }

  if (found) {
    for (j=i-1; j>0; j--) {
      Buffer[j].Fp = Buffer[j-1].Fp;
      Buffer[j].Offset = Buffer[j-1].Offset;
    }
    Buffer[0].Fp = fp;
    Buffer[0].Offset = offset;
  } else {
    tQS = QueueSize;
    QueueSize = ((tQS < BUF_SIZE) ? ++tQS : tQS);
    for (j = QueueSize-1; j > 0; j--) {
      Buffer[j].Fp = Buffer[j-1].Fp;
      Buffer[j].Offset = Buffer[j-1].Offset;
    }
    Buffer[0].Fp = fp;
    Buffer[0].Offset = offset;

    PageAccess++;
    /*    printf("PageAccess:%d\n",PageAccess);*/
  }

  return;
}


/***************************************************************************

* IO_DisplayPageAccess():
*   Routine to display the disk access
*
* IMPORTS:
*   long PageAccess
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void  IO_DisplayPageAccess(
) {
  (void)printf("Number of disk access :%16lu \n", (unsigned long)PageAccess);

  return;
}

void  IO_DisplayPageAccess_v2(
) {
  (void)printf("%16lu\t", (unsigned long)PageAccess);

  return;
}


/***************************************************************************

* IO_ResetPageAccess():
*   Routine to reset the PageAccess also refresh the buffer
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void IO_ResetPageAccess(void) {
  int i;

  PageAccess = 0;
  for (i=0; i < BUF_SIZE; i++){
    Buffer[i].Fp = 0;
    Buffer[i].Offset = 0;
  }
  QueueSize = 0;

  return;
}

void IO_ResetPageAccess_v2(void) {
  PageAccess = 0;
  return;
}

/***************************************************************************

* IOFileExist():
*   Routine to check whether a file exists
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TURE if file exists otherwise FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int IOFileExist(
		const char  *filename /* the filename of the file to check  */
		) {

  return(access(filename, F_OK)==0);

}

/***************************************************************************

* IOOpen():
*   Routine to open a file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   Returns the file handle of the opened file if it is opened successfully.
*   Otherwise, it returns NULL.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   the file is opened 
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
FILE* IOOpen(
	     const char  *filename,  /* the filename of the file to open */
	     const char *mode        /* the mode to open the file as     */
	     ) {
  FILE *fp;

  fp = fopen(filename, mode);
  if (fp == NULL)
    printf("Error opening file\n");

  return fp;
}

/***************************************************************************

* IOClose():
*   Routine to close a file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   Returns TRUE if the file us closed successfully,
*   Otherwise, it returns FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The file is closed
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int IOClose(
	    FILE  *fp /* the file handle of the file to be closed */
	    ) {
  int result;

  if (fclose(fp)==0)
    result = TRUE;
  else {
    printf("Error closing file\n");
    result = FALSE;
  }

  return result;
}
