/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


/*************************************************************************
 *
 * File : support.c
 *
 * Modules:
 * Functions support different mapping functions
 *	Dmax()
 *	Dmin()
 *	Z_Initialize()
 *	ZV()
 *	ZV1()
 *	getNextZvalue()
 *	getNextZvalueExit()
 *	IsViolation()
 *	Bit()
 *	DimValue()
 *	append()
 *	CompareViolation()
 *	pv()
 *	intersect()
 *	DeterminRange()
 *	MIN()
 *	MAX()
 *
 * Version 1.1
 * Coded by: Rui Zhang
 *************************************************************************/

#include <math.h>
#include "btree.h"
#include "support.h"

//iMinMax
int Dmax (RECORD_TYPE *pRcd)
{
  DATA_TYPE max = pRcd->fields[0];
  int dmax = 0, j;
  if (D > 1)
    {
      for (j = 1; j < D; j++)
	{
	  if (pRcd->fields[j] > max)
	    {
	      max = pRcd->fields[j];
	      dmax = j;
	    }
	}
    }
  return dmax;
}

int Dmin (RECORD_TYPE *pRcd)
{
  DATA_TYPE min = pRcd->fields[0];
  int dmin = 0, j;
  if (D > 1)
    {
      for (j = 1; j < D; j++)
	{
	  if (pRcd->fields[j] < min)
	    {
	      min = pRcd->fields[j];
	      dmin = j;
	    }
	}
    }
  return dmin;
}
//End iMinMax


//UB-tree
#define MAXPOS (D*ORDER-1)
#define dim(x) ((x)%D)

void Z_Initialize ()
{
  int i, ztypesize;

  zlimit = 0;
  zlimit = ~zlimit;

  ztypesize = sizeof (Z_TYPE) * 8;

  bitone = (Z_TYPE *) malloc (ztypesize * sizeof (Z_TYPE));
  bitzero = (Z_TYPE *) malloc (ztypesize * sizeof (Z_TYPE));

  for (i = 1, bitone[0] = 1; i < ztypesize; i++)
    bitone[i] = bitone[i-1] << 1;
  for (i = 0; i < ztypesize; i++)
    bitzero[i] = ~bitone[i];

  maxz = bitone[D*ORDER] - 1;
}

Z_TYPE ZV (RECORD_TYPE *pRcd)
{
  return ZV1 (pRcd); //data are normalized to [0,1]^d
}

//return z-value of a point pRcd, which is normalized to [0,1]
Z_TYPE ZV1 (RECORD_TYPE *pRcd)
{
  Z_TYPE zv = 0, bits, fieldj;
  int j, s;

  for (j = 0; j < D; j++)
    {
      fieldj = (Z_TYPE) floor ( (pRcd->fields[j]) * bitone[ORDER]);
      if (fieldj == (Z_TYPE) bitone[ORDER])
	fieldj--;
      for (s = 0; s < ORDER; s++)
	{
	  bits = fieldj & bitone[s]; //get bit s of field[j]
	  if (bits) //bit s of field[j] is 1
	    {
	      zv += bitone[s * D + j]; //set corresponding bit zv as 1
	      //printf("%llu\n",zv);
	    }
	}
    }
  return zv;
}

//This algorithm is described in [Integrating the UB-Tree into a Database System Kernel, VLDB2000].
//There seem to be a few errors on manipulating bit z, which have been corrected in the following
//function.
Z_TYPE getNextZvalue (Z_TYPE cur, Z_TYPE min, Z_TYPE max)
{
  short donegtMin[D], doneltMax[D];
  Z_TYPE temp;
  int p, v, j, gtMin[D], ltMax[D], outstepMin[D], outstepMax[D], outstep[D], conti, z;

  //bit scan, get the info of gtMin[j],ltMax[j],outstep[j]
  for (j = 0; j < D; j++)
    donegtMin[j] = doneltMax[j] = 0;

  for (p = MAXPOS; p >= 0; p--)
    {
      j = p%D;
      if (!donegtMin[j])
	{
	  if (Bit (cur, p) < Bit (min, p))
	    {
	      gtMin[j] = -1;
	      donegtMin[j] = 1;
	      outstepMin[j] = p;
	    }

	  else if (Bit (cur, p) > Bit (min, p))
	    {
	      gtMin[j] = p;
	      donegtMin[j] = 1;
	      outstepMin[j] = -1;
	    }

	  //else if (Bit(cur,p)==Bit(min,p) && p<D)
	  else if (p < D)
	    {
	      gtMin[j] = -1;
	      donegtMin[j] = 1;
	      outstepMin[j] = -1;
	    }
	}

      if (!doneltMax[j])
	{
	  if (Bit (cur, p) > Bit (max, p))
	    {
	      ltMax[j] = -1;
	      doneltMax[j] = 1;
	      outstepMax[j] = p;
	    }

	  else if (Bit (cur, p) < Bit (max, p))
	    {
	      ltMax[j] = p;
	      doneltMax[j] = 1;
	      outstepMax[j] = -1;
	    }

	  //else if (Bit(cur,p)==Bit(max,p) && p<D)
	  else if (p < D)
	    {
	      ltMax[j] = -1;
	      doneltMax[j] = 1;
	      outstepMax[j] = -1;
	    }
	}
    }

  for (j = 0; j < D; j++)
    {
      if (outstepMin[j] >= 0)
	outstep[j] = outstepMin[j];
      else if (outstepMax[j] >= 0)
	outstep[j] = outstepMax[j];
      else
	outstep[j] = -1;
    }

  v = outstep[0];//v is the highest bit position in cur that is part of a violation of the query box
  for (j = 1; j < D; j++)
    {
      if (outstep[j] > v)
	v = outstep[j];
    }

  if (v == -1)
    return cur;//cur is inside the query range

  //End bit scan, get the info of gtMin[j],ltMax[j],outstep[j]

  //find z
  conti = 1;
  z = -1;//z is the position greater or equal to v that can be safely set to 1
  for (p = v; p <= MAXPOS && conti; p++)
    {
      j = p % D;
      if (Bit (cur, p) == 0 && ltMax[j] >= p)
	{
	  z = p;
	  conti = 0;
	}
    }
  //End find z

  if (z == -1)
    return maxz;//cur is out of the query range
  else//manipulate cur into the next jump in point
    {
      cur = cur | bitone[z];
      for (p = 0, temp = zlimit; p<z; p++)
	temp -= bitone[p];
      cur = cur & temp;

      Z_TYPE cur_dim_z, min_dim_z;
      cur_dim_z = DimValue(cur, z);
      min_dim_z = DimValue(min, z);

      int setmin = 0;
      if (cur_dim_z < min_dim_z)
	setmin = 1; //this means the positions less than z in dim(z) should be set to minimum

      for (p = 0; p < z; p++)
	{
	  j = p % D;

	  if (j == z % D) //dim(z) has been changed when setting biz z as 1, so this dimension needs to be treated differently.
	    {
	      if (setmin && Bit (min, p))
		cur = cur | bitone[p];
	    }

	  else
	    {
	      if (gtMin[j] <= z && Bit (min,p))
		cur = cur | bitone[p];
	    }
	}
      //printf("%d: %d\n",oricur, cur);
      return cur;
    }

  getchar();
}

//input: zvalue zv and bit position p; output: the bit at position p
//note: position starts from right most bit, which is position 0.
int Bit (Z_TYPE zv, int p)
{
  if ( (zv & bitone[p]) > 0)
    return 1;
  else
    return 0;
}

//input: zvalue zv and dimension j; output: coordinate in dimension j
Z_TYPE DimValue (Z_TYPE zv, int j)
{
  Z_TYPE dv = 0;
  int p, step;
  j = j % D;
  for (p = j, step = 0; p < ORDER * D; p += D)
    {
      if (Bit (zv, p))
	dv = dv | bitone[step];
      step++;
    }
  return dv;
}

//This algorithm is adapted from Chapter 7 of
//[Towards a general-purpose, multidimensional index: Integration, Optimization, and
// Enhancement of UB-Trees, Thesis of Frank Ramsak, Institut fur Informatik TU chen, 2002].
// Again the algorithm given in the thesis has a few erros, which have been corrected
// in the following function.

Z_TYPE getNextZvalueExit (Z_TYPE cur, Z_TYPE min, Z_TYPE max)
{
  ValueNode *violations, vnode, *nextnode, *curnode;
  int min_cand, max_cand, nv, curv; //nv is the number of violations; curv is the current violation to process
  int p, i, j, dimv, v;
  int curbit, minbit, maxbit;
  int minfinished, maxfinished, vbetween=0, max_dimv_p;
  int *va; //violation array
  //int va[20];
  Z_TYPE njo, tmp, max_dimv;
	
  /////find violations////////

  violations = (ValueNode *) malloc (sizeof (ValueNode));
  violations->next = NULL;
  nv=0;
  //for each dimension check for violations
  for (i = 0; i < D; i++)
    {
      //initialize lists
      min_cand = max_cand = -1;
      minfinished = maxfinished = 0;
      //loop over all bits of the dimension, beginning with the highest valued one
      for (j = ORDER - 1; j >= 0; j--)
	{
	  //p=bp(i,j);
	  p = j * D + i;
	  curbit = Bit (cur, p);
	  minbit = Bit (min, p);
	  maxbit = Bit (max, p);
	  //looking for MIN violations
	  if (!minfinished)
	    {
	      if (curbit == minbit && minbit == 1 && min_cand == -1)
		//found MIN-violation in p
		append (violations, p, &nv);
	      else if (curbit == 1 && minbit == 0 && min_cand == -1)
		//found a MIN-violation candidate
		min_cand = p;
	      else if (minbit == 1 && min_cand >= 0)
		{
		  //the candidate is really a violation
		  append (violations, min_cand, &nv);
		  minfinished = 1;
		}
	    }//looking for MAXviolations
	  if (!maxfinished) 
	    {
	      if (curbit == maxbit && maxbit == 0 && max_cand == -1)
		//found a MAX-violation in p
		append (violations, p, &nv);
	      else if (curbit == 0 && maxbit == 1 && max_cand == -1)
		//found a MAX-violation candidate
		max_cand = p;
	      else if (maxbit == 0 && max_cand >= 0)
		{
		  //the candidate is really a violation
		  append (violations, max_cand, &nv);
		  maxfinished = 1;
		}
	    }
	}
    }

  //sort(violations);
  if ( (va = (int *) malloc (sizeof (int) *nv)) == NULL)
    {
      printf ("out of memory, use other sorting method!\n");
      exit (0);
    }

  //put the violations in an array and sort them
  nextnode = violations;
  for (i = 0; i < nv; i++)
    {
      va[i] = nextnode->v;
      nextnode = nextnode->next;
    }

  qsort(va, nv, sizeof(int), CompareViolation);

  //free the space
  for (nextnode = violations; nextnode != NULL;)
    {
      curnode = nextnode;
      nextnode = nextnode->next;
      free (curnode);
    }
  /////End find violations////////

  /////find exit/////////

  curv = 0;
  if (!nv)
    //the query box covers the complete universe
    return maxz;
  else
    {
      while (nv) 
	{
	  v = va[curv++];
	  nv--;
	  dimv = dim(v);
	  if (Bit (cur, v) == 0)
	    {
	      //v is MAX-violation
	      tmp = cur;

	      max_dimv = DimValue(max,dimv) + 1;
	      max_dimv_p = 0;

	      for (i = 0; i <= MAXPOS; i++)
		{
		  if (dim (i) == dimv)
		    {
		      if (Bit (max_dimv, max_dimv_p))
			tmp = tmp | bitone[i];
		      else
			tmp = tmp & bitzero[i];

		      max_dimv_p++;
		    }

		  else if (i < v)
		    tmp = tmp & bitzero[i];
		}

	      return (tmp - 1);
	    }

	  else
	    {
	      // v is MIN-violation

	      /*
		if(nv)
		{
		c=va[curv++];
		nv--;
		}
		else c=v;
	      */
	      for (i = v + 1; i <= MAXPOS; i++)
		{
		  if (Bit (cur, i) == 0)
		    {
		      if (dimv != dim (i) || vbetween) 
			{
			  njo = cur;
			  for (j = MAXPOS; j >= 0; j--)
			    {
			      if(j >= i) //njo[j]=cur[j]
				{
				  if (Bit(cur, j))
				    njo = njo | bitone[j];
				  else
				    njo = njo & bitzero[j];
				}
			      else
				njo = njo | bitone[j];
			    }
			  return njo;
			}//the zero-bit is also MAX-violation handle this case

		      //else if(c == i)
		      else if( curv = IsViolation(i, va, curv, nv))
			{
			  /*
			  //push(violations,i);
			  curv--;
			  va[curv]=i;
			  nv++;
			  break;
			  */

			  break;
			}
		    }

		  //else if(i == c && dimv != dim(i))
		  else if (dimv != dim(i) && IsViolation(i, va, curv, nv))
		    {
		      vbetween = 1;

		      /*
			if(nv)
			{
			c=va[curv++];
			nv--;
			}
		      */
		    }
		}//if there no zero bit that can be safely set then NJO=max

	      if (i > MAXPOS)
		return max;
	    }
	}
    }
  /////End find exit/////////

  free(va);
}

int IsViolation (int p, int va[], int curv, int nv)
{
  int i;
  for (i = curv; i < curv + nv; i++)
    {
      if (p == va[i])
	return i;
    }
  return 0;
}

void append (ValueNode *v, int a, int *nv)
{
  ValueNode *curvn;
  if (v->next == NULL)
    {
      curvn = (ValueNode *) malloc (sizeof (ValueNode));
      curvn->next = NULL;
      v->next = curvn;
      v->v = a;
      (*nv)++;
    }

  else
    append(v->next, a, nv);
}

int CompareViolation (const void *arg1, const void *arg2)
{
  if (* (int *)arg1 < * (int *)arg2)
    return -1;
  else if (* (int *)arg1 > * (int *)arg2)
    return 1;
  else
    return 0;	
}

//End UB-tree


//Pyramid technique

KEYTYPE pv (DATA_TYPE * vector) //pyramid value
{
  pyrval p = {-1, -1.0};
  int j;
  for (j = 0; j < D; j++)
    if (vector[j] <= 0.5)
      {
	if (p.height < 0.5 - vector [j])
	  {
	    p.height = 0.5 - vector [j];
	    p.pyramid = j;
	  }
      }
    else
      {
	if (p.height < vector [j] - 0.5)
	  {
	    p.height = vector [j] - 0.5;
	    p.pyramid = j + D;
	  }
    }
  return p.height + p.pyramid;
}

int intersect (RANGE_TYPE *pRg, int i, float *phlow, float *phhigh)
{
  int j;
  RANGE_TYPE pRg1;
  for (j = 0; j < D; j++)
    {
      pRg1.lower[j] = pRg->lower[j] - 0.5;
      pRg1.higher[j] = pRg->higher[j] - 0.5;
    }

  if (i < D)
    {
      //for(j=0;j<D && qw[i][0]<=-MIN(qw[j]);j++);
      for (j = 0; j < D && pRg1.lower[i] <= -MIN(pRg1.lower[j], pRg1.higher[j]); j++);
      if (j == D)
	{
	  //if (qw[i][1]>0) qw[i][1]=0;
	  if (pRg1.higher[i] > 0)
	    pRg1.higher[i] = 0;
	  DeterminRange(&pRg1, i, phlow, phhigh);
	  return 1;
	}
      else
	return 0;
    }

  if (D - 1 < i)
    {
      //for(j=0;j<D && qw[i-D][1]>=MIN(qw[j]);j++);
      for (j = 0; j < D && pRg1.higher[i - D] >= MIN(pRg1.lower[j], pRg1.higher[j]); j++);
      if (j == D)
	{
	  //if (qw[i-D][0]<0) qw[i-D][0]=0;
	  if (pRg1.lower[i - D] < 0)
	    pRg1.lower[i - D] = 0;
	  DeterminRange (&pRg1, i - D, phlow, phhigh);
	  return 1;
	}
      else
	return 0;
    }
}

void DeterminRange (RANGE_TYPE *pRg, int i, float *phlow, float *phhigh)
{
  int j, k;
  float qtemp, qtemp1;
  //*phhigh=MAX(qw[i]);
  *phhigh = MAX (pRg->lower[i], pRg->higher[i]);
  //for(j=0;j<D && qw[j][0]<=0 && 0<=qw[j][1]; j++);
  for (j = 0; j < D && pRg->lower[j] <= 0 && 0 <= pRg->higher[j]; j++);
  if (j == D)
    *phlow = 0;
  else
    {
      *phlow = 0.0;
      for (k = 0; k < D; k++)
	{
	  //qtemp=MIN(qw[k])>MIN(qw[i])? MIN(qw[k]):MIN(qw[i]);
	  qtemp1 = MIN (pRg->lower[k], pRg->higher[k]);
	  qtemp = MIN (pRg->lower[i], pRg->higher[i]);
	  if(qtemp1 > qtemp)
	    qtemp = qtemp1;
	  if(qtemp > *phlow)
	    *phlow = qtemp;
	}
    }
}

float MIN(float f0, float f1)
{
  if (f0 * f1 <= 0)
    return 0;
  else
    return fabs(f0) < fabs(f1) ? fabs(f0) : fabs(f1);
}

float MAX(float f0, float f1)
{
  return fabs(f0) > fabs(f1) ? fabs(f0) : fabs(f1);
}

//End Pyramid technique
